from __future__ import annotations
from typing import Union, Optional, List

from ply import lex

# TODO: add -d/--debug to make-spec
DEBUG = False


def _dbg_print(msg):
    if DEBUG:
        print(msg)


# TODO: probably want to refactor lexer to make it look like lexer.py.


class Lexer:
    tokens = (
        "NEWLINE",
        "WS",
        "LBRACKET",
        "RBRACKET",
        "LPAREN",
        "RPAREN",
        "LBRACE",
        "RBRACE",
        "PIPE",
        "SWITCH",
        "VALUE",
    )

    def t_NEWLINE(self, t):
        r"\n"
        return t

    def t_WS(self, t):
        r"\s+"
        pass

    def t_LBRACKET(self, t):
        r"\["
        return t

    def t_RBRACKET(self, t):
        r"\]"
        return t

    def t_LPAREN(self, t):
        r"\("
        return t

    def t_RPAREN(self, t):
        r"\)"
        return t

    def t_LBRACE(self, t):
        r"\{"
        return t

    def t_RBRACE(self, t):
        r"\}"
        return t

    def t_PIPE(self, t):
        r"\|"
        return t

    def t_SWITCH(self, t):
        r"(-[^\[\]\s\|\(\)]+|>+)"
        return t

    def t_VALUE(self, t):
        r"[^\[\]\s\|\(\)\{\}]+"
        return t

    def t_error(self, t):
        print("Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    def __init__(self, data):
        self.lexer = lex.lex(object=self)
        self.lexer.input(data)
        self.current = self.lexer.token()
        self.next_tok = self.lexer.token()

    def next(self):
        current = self.current
        self.current = self.next_tok
        self.next_tok = self.lexer.token()
        return current

    def token(self):
        return self.next()


ArgType = Union["Positional", "Switch", "OptionalArg", "ExclusiveArgs"]
ValueType = Union["AnyVal", "ExclusiveVals", "Literal", "ListVal"]


class Command:
    def __init__(self, name: str, args: Optional[List[ArgType]] = None):
        self.name = name
        self.args = []
        if args is not None:
            self.args = args

    def __str__(self):
        return f"{self.name} {' '.join(map(str, self.args))}"

    def __eq__(self, other):
        if not isinstance(other, Command):
            return False
        return (self.name == other.name) and (self.args == other.args)


class AnyVal:
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if not isinstance(other, AnyVal):
            return False
        return self.name == other.name


class Positional:
    def __init__(self, value: ValueType):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        if not isinstance(other, Positional):
            return False
        return self.value == other.value


class Switch:
    def __init__(self, name, value: Optional[ValueType] = None):
        self.name = name
        self.value = value

    def __str__(self):
        s = f"{self.name}"
        if self.value is not None:
            s += f" {self.value}"
        return s

    def __eq__(self, other):
        if not isinstance(other, Switch):
            return False
        return (self.name == other.name) and (self.value == other.value)


class ListVal:
    def __init__(self, items: Optional[List[str]] = None):
        self.items = []
        if items is not None:
            self.items = items

    def __str__(self):
        return "{" + " ".join(self.items) + "}"

    def __eq__(self, other):
        if not isinstance(other, ListVal):
            return False
        return self.items == other.items


class OptionalArg:
    def __init__(self, child: ArgType):
        self.child = child

    def __str__(self):
        return f"[{self.child}]"

    def __eq__(self, other):
        if not isinstance(other, OptionalArg):
            return False
        return self.child == other.child


class ExclusiveArgs:
    def __init__(self, choices: Optional[List[ArgType]] = None):
        self.choices = []
        if choices is not None:
            self.choices = choices

    def __str__(self):
        return "(" + " | ".join(map(str, self.choices)) + ")"

    def __eq__(self, other):
        if not isinstance(other, ExclusiveArgs):
            return False
        return self.choices == other.choices


class ExclusiveVals:
    def __init__(self, choices: Optional[List[Union[AnyVal, ListVal, Literal]]] = None):
        self.choices = []
        if choices is not None:
            self.choices = choices

    def __str__(self):
        return "(" + " | ".join(map(str, self.choices)) + ")"

    def __eq__(self, other):
        if not isinstance(other, ExclusiveVals):
            return False
        return self.choices == other.choices


class Literal:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return f'"{self.value}"'

    def __eq__(self, other):
        if not isinstance(other, Literal):
            return False
        return self.value == other.value


def parse_help_entry(entry: str) -> Command:
    lexer = Lexer(entry)
    return parse_command(lexer)


def spec_from_help_entry(entry: str) -> dict:
    command = parse_help_entry(entry)
    return {command.name: make_spec(command)}


def parse_command(lexer: Lexer) -> Command:
    _dbg_print(f"parse_command({lexer.current})")
    tok = lexer.token()

    assert tok.type == "VALUE", f"Expected value, got {tok.type}"
    command = Command(tok.value)

    while lexer.current is not None and lexer.current.type != "NEWLINE":
        arg = parse_arg(lexer)
        command.args.append(arg)
    lexer.token()  # munch newline

    return command


def parse_arg(lexer: Lexer) -> ArgType:
    _dbg_print(f"parse_arg({lexer.current})")
    tok = lexer.current

    arg: Optional[ArgType] = None
    if tok.type == "LPAREN":
        lexer.token()
        arg = parse_arg(lexer)
        tok = lexer.token()
        assert tok.type == "RPAREN", f"expected closing paren, got {tok.type}"
        return arg
    elif tok.type == "LBRACKET":
        lexer.token()
        arg = OptionalArg(parse_arg(lexer))
        assert lexer.token().type == "RBRACKET", "expected closing bracket"
    elif tok.type == "SWITCH":
        lexer.token()
        value = None
        if lexer.current is not None and lexer.current.type in {
            "VALUE",
            "LPAREN",
            "LBRACE",
        }:
            value = parse_value(lexer)
        arg = Switch(tok.value, value)
    elif tok.type in {"VALUE", "LBRACE"}:
        arg = Positional(parse_value(lexer))

    assert arg is not None, f"Unexpected token {tok.type}"

    if lexer.current is not None and lexer.current.type == "PIPE":
        lexer.token()  # munch pipe

        exclusive = ExclusiveArgs()
        exclusive.choices.append(arg)

        next = parse_arg(lexer)
        if isinstance(next, ExclusiveArgs):
            # flatten this
            exclusive.choices.extend(next.choices)
        else:
            exclusive.choices.append(next)

        arg = exclusive

    return arg


def parse_value(lexer: Lexer) -> ValueType:
    _dbg_print(f"parse_value({lexer.current})")
    tok = lexer.next()
    val: ValueType
    if tok.type == "LPAREN":
        val = parse_value(lexer)
        assert lexer.token().type == "RPAREN"
        return val

    if tok.type == "LBRACE":
        val = ListVal()
        while lexer.current.type != "RBRACE":
            val.items.append(lexer.current.value)
            lexer.token()
        assert lexer.token().type == "RBRACE"
    else:
        assert tok.type == "VALUE", tok.type
        val = AnyVal(tok.value)

    if (
        lexer.current is not None
        and lexer.current.type == "PIPE"
        and lexer.next_tok is not None
        and lexer.next_tok.type in {"VALUE", "LBRACE"}
    ):
        lexer.next()  # munch pipe
        choices: List[Union[AnyVal, ListVal, Literal]] = [val]

        next = parse_value(lexer)
        if isinstance(next, ExclusiveVals):
            # flatten this
            choices.extend(next.choices)
        else:
            choices.append(next)

        make_literal = [isinstance(c, (AnyVal, Literal)) for c in choices].count(
            True
        ) > 1

        exclusive = ExclusiveVals()
        for choice in choices:
            if make_literal and isinstance(choice, AnyVal):
                exclusive.choices.append(Literal(choice.name))
            else:
                exclusive.choices.append(choice)

        return exclusive
    else:
        return val


def make_spec(command: Command) -> dict:
    spec = {"": {"min": 0, "max": 0}}

    for arg in command.args:
        if isinstance(arg, OptionalArg):
            _process_arg(arg.child, spec, optional=True)
        else:
            _process_arg(arg, spec, optional=False)

    for redirect in (">", ">>"):
        if redirect not in spec:
            spec[redirect] = {"required": False, "value": True, "repeated": False}

    return spec


def _process_arg(arg: ArgType, spec: dict, optional=False):
    if isinstance(arg, Positional):
        if not optional:
            spec[""]["min"] += 1
        spec[""]["max"] += 1
    elif isinstance(arg, Switch):
        spec[arg.name] = {
            "required": not optional,
            "value": arg.value is not None,
            "repeated": False,
        }
    elif isinstance(arg, ExclusiveArgs):
        has_positional = False
        for subarg in arg.choices:
            # bit of a hack, right now everything under ExclusiveArgs is treated
            # as effectively optional, so just pop into the child of the
            # Optional
            if isinstance(subarg, OptionalArg):
                subarg = subarg.child

            assert not isinstance(
                subarg, (ExclusiveArgs, OptionalArg)
            ), "Invalid argument type inside exclusive args"

            if isinstance(subarg, Switch):
                # TODO: make this stricter w/ exclusive field
                spec[subarg.name] = {
                    "required": False,
                    "value": subarg.value is not None,
                    "repeated": False,
                }
            elif isinstance(subarg, Positional):
                has_positional = True
        if has_positional:
            spec[""]["max"] += 1



from tclint.plugins.openroad.help_parser import (
    parse_help_entry,
    make_spec,
    Command,
    Switch,
    OptionalArg,
    Positional,
    AnyVal,
    ExclusiveArgs,
    ExclusiveVals,
    Literal,
    ListVal,
)


def test_basic_help():
    tree = parse_help_entry("command -switch [optional] positional")
    expected = Command(
        "command",
        [
            Switch("-switch"),
            OptionalArg(Positional(AnyVal("optional"))),
            Positional(AnyVal("positional")),
        ],
    )
    assert tree == expected, f"{tree} != {expected}"


def test_exclusive():
    tree = parse_help_entry("estimate_parasitics -placement|-global_routing")
    expected = Command(
        "estimate_parasitics",
        [
            ExclusiveArgs([
                Switch("-placement"),
                Switch("-global_routing"),
            ])
        ],
    )
    assert tree == expected, f"{tree} != {expected}"


def test_exclusive_mixed_keys():
    tree = parse_help_entry("command [-foo foo | -bar | -baz baz]")
    expected = Command(
        "command",
        [
            OptionalArg(
                ExclusiveArgs([
                    Switch("-foo", AnyVal("foo")),
                    Switch("-bar"),
                    Switch("-baz", AnyVal("baz")),
                ])
            )
        ],
    )
    assert tree == expected, f"{tree} != {expected}"


def test_exclusive_literal():
    tree = parse_help_entry("command -switch a|b|c")
    expected = Command(
        "command",
        [Switch("-switch", ExclusiveVals([Literal("a"), Literal("b"), Literal("c")]))],
    )
    assert tree == expected, f"{tree} != {expected}"


def test_exclusive_literal_parens_spaces():
    tree = parse_help_entry("command -switch ( a|b |c)")
    expected = Command(
        "command",
        [Switch("-switch", ExclusiveVals([Literal("a"), Literal("b"), Literal("c")]))],
    )
    assert tree == expected, f"{tree} != {expected}"


def test_careted_metavar():
    tree = parse_help_entry("define_pdn_grid [-name <name>]")
    expected = Command(
        "define_pdn_grid", [OptionalArg(Switch("-name", AnyVal("<name>")))]
    )
    assert tree == expected, f"{tree} != {expected}"


def test_list():
    tree = parse_help_entry("density_fill -area {lx ly ux uy}")
    expected = Command(
        "density_fill", [Switch("-area", ListVal(["lx", "ly", "ux", "uy"]))]
    )
    assert tree == expected, f"{tree} != {expected}"


def test_exclusive_any_list():
    tree = parse_help_entry("improve_placement -max_displacement disp|{disp_x disp_y}")
    expected = Command(
        "improve_placement",
        [
            Switch(
                "-max_displacement",
                ExclusiveVals([AnyVal("disp"), ListVal(["disp_x", "disp_y"])]),
            )
        ],
    )
    assert tree == expected, f"{tree} != {expected}"


def test_exclusive_list_any():
    tree = parse_help_entry("improve_placement -max_displacement {disp_x disp_y}|disp")
    expected = Command(
        "improve_placement",
        [
            Switch(
                "-max_displacement",
                ExclusiveVals([ListVal(["disp_x", "disp_y"]), AnyVal("disp")]),
            )
        ],
    )
    assert tree == expected, f"{tree} != {expected}"


def test_exclusive_optionals():
    tree = parse_help_entry("command [-foo] | [-bar]")
    spec = make_spec(tree)

    assert "-foo" in spec and spec["-foo"] == {
        "repeated": False,
        "required": False,
        "value": False,
    }
    assert "-bar" in spec and spec["-bar"] == {
        "repeated": False,
        "required": False,
        "value": False,
    }


def test_parenthesized_args():
    help = "command [(-foo|-bar)]"
    tree = parse_help_entry(help)
    expected = Command(
        "command", [OptionalArg(ExclusiveArgs([Switch("-foo"), Switch("-bar")]))]
    )
    assert tree == expected, f"{tree} != {expected}"



import pytest

from tclint.commands import CommandArgError
from tclint.plugins.openroad.openroad import check_arg_spec
from tclint.syntax_tree import BareWord, ArgExpansion, VarSub


def test_repeated_switch_allowed():
    args = [BareWord("-abc"), BareWord("-abc")]
    spec = {
        "": {"min": 0, "max": 0},
        "-abc": {"required": False, "value": False, "repeated": True},
    }
    assert check_arg_spec("command", spec)(args, None) is None


def test_repeated_switch_not_allowed():
    args = [BareWord("-abc"), BareWord("-abc")]
    spec = {
        "": {"min": 0, "max": 0},
        "-abc": {"required": False, "value": False, "repeated": False},
    }
    with pytest.raises(CommandArgError):
        check_arg_spec("command", spec)(args, None)


def test_positional_count_too_many():
    args = [BareWord("foo")]
    spec = {"": {"min": 0, "max": 0}}
    with pytest.raises(CommandArgError):
        check_arg_spec("command", spec)(args, None)


def test_positional_count_too_few():
    args = [BareWord("foo")]
    spec = {"": {"min": 2, "max": 2}}
    with pytest.raises(CommandArgError):
        check_arg_spec("command", spec)(args, None)


def test_positional_count_unlimited():
    args = [BareWord("foo")] * 10
    spec = {"": {"min": 1, "max": None}}
    check_arg_spec("command", spec)(args, None)


def test_positional_count_arg_expansion():
    args = [BareWord("foo"), ArgExpansion(VarSub("foo"))]
    spec = {"": {"min": 3, "max": None}}
    check_arg_spec("command", spec)(args, None)



import argparse
import logging
from pathlib import Path
from typing import List
import uuid

from lsprotocol import types as lsp

from pygls.server import LanguageServer
from pygls.workspace import TextDocument
from pygls.uris import to_fs_path

from tclint import main as tclint
from tclint.config import get_config, DEFAULT_CONFIGS, Config, ConfigError
from tclint.format import Formatter, FormatterOpts
from tclint.lexer import TclSyntaxError
from tclint.parser import Parser

try:
    from tclint._version import __version__  # type: ignore
except ModuleNotFoundError:
    __version__ = "(unknown version)"


DIAGNOSTIC_SOURCE = "tclint"


def lint(source, config, path):
    diagnostics = []

    try:
        violations = tclint.lint(source, config, path)
    except TclSyntaxError as e:
        return [
            lsp.Diagnostic(
                message=str(e),
                severity=lsp.DiagnosticSeverity.Error,
                range=lsp.Range(
                    start=lsp.Position(e.start[0] - 1, e.start[1] - 1),
                    end=lsp.Position(e.end[0] - 1, e.end[1] - 1),
                ),
                code="syntax error",
                source=DIAGNOSTIC_SOURCE,
            )
        ]

    for violation in violations:
        message = violation.message
        severity = lsp.DiagnosticSeverity.Warning
        start = lsp.Position(
            line=violation.start[0] - 1, character=violation.start[1] - 1
        )
        end = lsp.Position(line=violation.end[0] - 1, character=violation.end[1] - 1)

        diagnostics.append(
            lsp.Diagnostic(
                message=message,
                severity=severity,
                range=lsp.Range(
                    start=start,
                    end=end,
                ),
                code=violation.id,
                source=DIAGNOSTIC_SOURCE,
            )
        )

    return diagnostics


class TclspServer(LanguageServer):
    """Main server class. Implements pull diagnostics using a method adapted from
    https://pygls.readthedocs.io/en/latest/examples/pull-diagnostics.html."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.diagnostics = {}
        # maps root path -> RunConfig, since there may be more than one workspace.
        self.configs = {}
        self.client_supports_refresh = False

    def get_roots(self) -> List[Path]:
        roots = []
        for uri in self.workspace.folders.keys():
            path = to_fs_path(uri)
            if path is not None:
                roots.append(Path(path))

        if len(roots) > 0:
            return roots

        if self.workspace.root_path is not None:
            roots.append(Path(self.workspace.root_path))

        return roots

    def load_configs(self):
        self.configs = {}
        for root in self.get_roots():
            config = None
            try:
                config = get_config(None, root)
            except ConfigError as e:
                self.show_message(f"Error loading config file: {e}")

            if config is not None:
                for other in self.configs.keys():
                    if root.is_relative_to(other) or other.is_relative_to(root):
                        self.show_message(
                            f"Warning: found configs in overlapping workspaces: {root},"
                            f" {other}. It's undefined which will apply."
                        )

                self.configs[root] = config

    def get_config(self, path: Path) -> Config:
        matching_config = Config()
        for root, config in self.configs.items():
            if path.is_relative_to(root):
                matching_config = config.get_for_path(path)
                break
        return matching_config

    def parse(self, document: TextDocument):
        path = Path(document.path)
        config = self.get_config(path)

        # `None` sentinel ensures that `diagnostics` gets updated if the URI is not
        # present.
        _, previous = self.diagnostics.get(document.uri, (0, None))
        diagnostics = lint(document.source, config, path)

        # Only update if the list has changed
        if previous != diagnostics:
            self.diagnostics[document.uri] = (document.version, diagnostics)

    def format(self, document: TextDocument, options: lsp.FormattingOptions):
        path = Path(document.path)
        config = self.get_config(path)

        parser = Parser()

        if config.style_indent is None:
            indent = "\t" if not options.insert_spaces else " " * options.tab_size
        else:
            indent = config.get_indent()

        formatter = Formatter(
            FormatterOpts(
                indent=indent,
                spaces_in_braces=config.style_spaces_in_braces,
                max_blank_lines=config.style_max_blank_lines,
                indent_namespace_eval=config.style_indent_namespace_eval,
            )
        )
        return formatter.format_top(document.source, parser)


server = TclspServer("tclsp", __version__)


@server.feature(lsp.TEXT_DOCUMENT_DID_OPEN)
def did_open(ls: TclspServer, params: lsp.DidOpenTextDocumentParams):
    """Parse each document when it is opened"""
    logging.debug("Received %s: %s", lsp.TEXT_DOCUMENT_DID_OPEN, params)
    doc = ls.workspace.get_text_document(params.text_document.uri)
    ls.parse(doc)


@server.feature(lsp.TEXT_DOCUMENT_DID_CHANGE)
def did_change(ls: TclspServer, params: lsp.DidOpenTextDocumentParams):
    """Parse each document when it is changed"""
    logging.debug("Received %s: %s", lsp.TEXT_DOCUMENT_DID_CHANGE, params)
    doc = ls.workspace.get_text_document(params.text_document.uri)
    ls.parse(doc)


@server.feature(
    lsp.TEXT_DOCUMENT_DIAGNOSTIC,
    lsp.DiagnosticOptions(
        identifier="pull-diagnostics",
        inter_file_dependencies=False,
        # We could support workspace diagnostics, although an implementation based on
        # the pygls tutorial seems to add client-server noise for no benefit (it ends up
        # replying to a frequent workspace diagnostics request with "unchanged"
        # messages).
        workspace_diagnostics=False,
    ),
)
def document_diagnostic(ls: TclspServer, params: lsp.DocumentDiagnosticParams):
    """Return diagnostics for the requested document"""
    logging.debug("Received %s: %s", lsp.TEXT_DOCUMENT_DIAGNOSTIC, params)

    was_cached = True
    if (uri := params.text_document.uri) not in ls.diagnostics:
        was_cached = False
        doc = ls.workspace.get_text_document(uri)
        ls.parse(doc)

    version, diagnostics = ls.diagnostics[uri]
    result_id = f"{uri}@{version}"

    if was_cached and result_id == params.previous_result_id:
        return lsp.UnchangedDocumentDiagnosticReport(result_id)

    return lsp.FullDocumentDiagnosticReport(items=diagnostics, result_id=result_id)


@server.feature(lsp.WORKSPACE_DID_CHANGE_WATCHED_FILES)
def change_watched_files(ls: TclspServer, params: lsp.DidChangeWatchedFilesParams):
    logging.debug("Received %s: %s", lsp.WORKSPACE_DID_CHANGE_WATCHED_FILES, params)

    # Clear diagnostics cache so they get recalculated when requested
    ls.diagnostics = {}

    ls.load_configs()
    if ls.client_supports_refresh:
        ls.lsp.send_request(lsp.WORKSPACE_DIAGNOSTIC_REFRESH, None)


@server.feature(lsp.TEXT_DOCUMENT_FORMATTING)
def format_document(ls: TclspServer, params: lsp.DocumentFormattingParams):
    """Format the entire document"""
    doc = ls.workspace.get_text_document(params.text_document.uri)

    source = doc.source
    start = lsp.Position(line=0, character=0)
    last_line = source.rsplit("\n", 1)[-1]
    end = lsp.Position(line=source.count("\n"), character=len(last_line))

    formatted = ls.format(doc, params.options)
    return [
        lsp.TextEdit(
            range=lsp.Range(start=start, end=end),
            new_text=formatted,
        )
    ]


@server.feature(lsp.INITIALIZED)
def init(ls: TclspServer, params: lsp.InitializeParams):
    """Registers file watchers on config filenames so that we can reload configs and
    refresh diagnostics if they've changed.

    Based on code snippet in
    https://github.com/openlawlibrary/pygls/issues/376#issuecomment-1717656614.
    """
    capabilities = ls.client_capabilities.workspace

    try:
        ls.client_supports_refresh = (
            capabilities.diagnostics.refresh_support  # type: ignore[union-attr]
        )
    except AttributeError:
        ls.client_supports_refresh = False

    try:
        client_supports_watched_files_registration = (
            capabilities.did_change_watched_files.dynamic_registration  # type: ignore[union-attr] # noqa: E501
        )
    except AttributeError:
        client_supports_watched_files_registration = False

    if client_supports_watched_files_registration:
        watchers = []
        for filename in (*DEFAULT_CONFIGS, "pyproject.toml"):
            pattern = f"**/{filename}"
            watchers.append(lsp.FileSystemWatcher(glob_pattern=pattern))

        ls.register_capability(
            lsp.RegistrationParams(
                registrations=[
                    lsp.Registration(
                        id=str(uuid.uuid4()),
                        method=lsp.WORKSPACE_DID_CHANGE_WATCHED_FILES,
                        register_options=lsp.DidChangeWatchedFilesRegistrationOptions(
                            watchers=watchers
                        ),
                    )
                ]
            )
        )

    ls.load_configs()


def main():
    parser = argparse.ArgumentParser("tclsp")
    log_levels = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
    }
    parser.add_argument(
        "-l",
        "--log-level",
        default="info",
        type=lambda x: x.lower(),
        help="set the log level. defaults to info",
        choices=log_levels.keys(),
    )
    args = parser.parse_args()
    logging.basicConfig(level=log_levels[args.log_level], format="%(message)s")

    server.start_io()


if __name__ == "__main__":
    main()



"""Parse-time handling of Tcl's builtin commands.

Note that the following commands are not currently supported. If support for any of
these would be helpful for your use case, please file an issue.

- Anything related to TclOO:
  - https://www.tcl.tk/man/tcl/TclCmd/my.html
  - https://www.tcl.tk/man/tcl/TclCmd/next.html
  - https://www.tcl.tk/man/tcl/TclCmd/class.html
  - https://www.tcl.tk/man/tcl/TclCmd/copy.html
  - https://www.tcl.tk/man/tcl/TclCmd/define.html
  - https://www.tcl.tk/man/tcl/TclCmd/object.html
  - https://www.tcl.tk/man/tcl/TclCmd/self.html

- Things that are imported via `package require`
  - https://www.tcl.tk/man/tcl/TclCmd/dde.html
  - https://www.tcl.tk/man/tcl/TclCmd/http.html
  - https://www.tcl.tk/man/tcl/TclCmd/msgcat.html
  - https://www.tcl.tk/man/tcl/TclCmd/platform.html
  - https://www.tcl.tk/man/tcl/TclCmd/platform_shell.html
  - https://www.tcl.tk/man/tcl/TclCmd/transchan.html
  - https://www.tcl.tk/man/tcl/TclCmd/tcltest.html

- Tcl library commands: https://www.tcl.tk/man/tcl/TclCmd/library.html

- The "unknown" command: https://www.tcl.tk/man/tcl/TclCmd/unknown.html

- Math ops:
  - https://www.tcl.tk/man/tcl/TclCmd/mathfunc.html
  - https://www.tcl.tk/man/tcl/TclCmd/mathop.html
"""

from tclint.commands.utils import (
    CommandArgError,
    check_count,
    subcommands,
    eval,
)
from tclint.syntax_tree import BareWord


def _check_code(arg):
    """Check 'code' argument used by return and try."""

    val = arg.contents
    if val is None:
        return

    try:
        int(val)
    except ValueError:
        pass
    else:
        return

    if val in {"ok", "error", "return", "break", "continue"}:
        return

    raise CommandArgError(
        f"got {val}, expected one of ok, error, return, break, continue, or an integer"
    )


def _after(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/after.html

    script_arg = []
    if len(args) > 1:
        script_arg = eval(args[1:], parser, "after")

    return args[0:1] + script_arg


def _after_cancel(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/after.html
    check_count("after cancel", 1, None)

    # TODO: raise warning about not checking code

    return None


def _after_idle(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/after.html
    return eval(args, parser, "after idle")


def _apply(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/apply.html
    if len(args) < 1:
        raise CommandArgError(
            f"not enough args to apply: got {len(args)}, expected at least 1"
        )

    func_list = parser.parse_list(args[0])
    list_len = len(func_list.children)
    if list_len < 2 or list_len > 3:
        raise CommandArgError(
            f"Invalid first argument to apply: got list of {list_len} elements,"
            " expected 2 or 3"
        )

    body = parser.parse_script(func_list.children[1])
    func_list.children[1] = body

    return [func_list] + args[1:]


def _catch(args, parser):
    if len(args) < 1:
        raise CommandArgError(
            f"not enough args to catch: got {len(args)}, expected at least 1"
        )
    if len(args) > 3:
        raise CommandArgError(
            f"too many args to catch: got {len(args)}, expected no more than 3"
        )

    return [parser.parse_script(args[0])] + args[1:]


def _dict_filter(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/dict.html#M8

    if len(args) < 2:
        raise CommandArgError(
            f"not enough args to 'dict filter': got {len(args)}, expected at least 2"
        )

    if args[1].contents not in {"key", "script", "value"}:
        raise CommandArgError(
            "invalid argument to 'dict filter': expected filter type to be one of key,"
            " script, or value"
        )

    if args[1].contents == "script":
        kv_pair = parser.parse_list(args[2])
        list_len = len(kv_pair.children)
        if len(kv_pair.children) != 2:
            raise CommandArgError(
                "invalid argument to 'dict filter': expected list of 2 elements in"
                f" second-to-last argument, got {list_len}"
            )
        return args[0:2] + [kv_pair, parser.parse_script(args[3])]

    return None


def _dict_map_for(cmd):
    def check(args, parser):
        if len(args) != 3:
            raise CommandArgError(
                f"wrong # of args to '{cmd}': got {len(args)}, expected 3"
            )

        # TODO: might be worth checking that arg[0] is a pair?

        return args[0:2] + [parser.parse_script(args[2])]

    return check


def _dict_update(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/dict.html#M25

    if len(args) < 4:
        raise CommandArgError(
            f"not enough args to 'dict update': got {len(args)}, expected at least 4"
        )

    if len(args) % 2 != 0:
        raise CommandArgError(
            "invalid # of args to 'dict update': expected an even number"
        )

    return args[0:-1] + [parser.parse_script(args[-1])]


def _dict_with(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/dict.html#M27

    if len(args) < 2:
        raise CommandArgError(
            f"not enough args to 'dict with': got {len(args)}, expected at least 2"
        )

    return args[0:-1] + [parser.parse_script(args[-1])]


def _eval(args, parser):
    return eval(args, parser, "eval")


def _expr(args, parser):
    if len(args) == 0:
        raise CommandArgError("not enough args to 'expr': got 0, expected at least 1")

    # Handle single argument consisting of BareWord, BracedWord, or concrete QuotedWord.
    if len(args) == 1 and args[0].contents is not None:
        # this method will handle the `node.contents is None` case fine, but
        # will throw an error. We'll instead pass thru silently, since that error
        # will be caught by a separate lint check.
        return [parser.parse_expression(args[0])]

    # Handle multiple BareWord arguments. Non-BareWords are hard to handle in this case,
    # since we need to pop the contents out of quoted or braced words, but then we have
    # no way of storing the original info about these words in the syntax tree.
    contents = ""
    last_pos = args[0].pos
    for arg in args:
        if not isinstance(arg, BareWord):
            return None

        if arg.pos[0] != last_pos[0]:
            contents += "\n" * (arg.pos[0] - last_pos[0])
            contents += " " * (arg.pos[1] - 1)
        else:
            contents += " " * (arg.pos[1] - last_pos[1])
        contents += arg.contents
        last_pos = arg.end_pos

    node = BareWord(contents, pos=args[0].pos, end_pos=args[-1].end_pos)
    return [parser.parse_expression(node)]


def _fileevent(args, parser):
    # ref: https://www.tcl.tk/man/tcl8.4/TclCmd/fileevent.html
    # TODO: implement
    raise CommandArgError(
        "argument parsing for 'fileevent' not implemented, script argument will not be"
        " checked for violations"
    )


def _for(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/for.html
    if len(args) != 4:
        raise CommandArgError(f"wrong # of args to for: got {len(args)}, expected 4")

    return [
        parser.parse_script(args[0]),
        parser.parse_expression(args[1]),
        parser.parse_script(args[2]),
        parser.parse_script(args[3]),
    ]


def _foreach(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/foreach.html
    if len(args) < 3:
        raise CommandArgError(
            f"insufficient args to foreach: got {len(args)}, expected at least 3"
        )

    # last argument is script body
    return args[0:-1] + [parser.parse_script(args[-1])]


def _if(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/if.html
    # TODO: make arg checking strict

    new_args = []

    new_args.append(parser.parse_expression(args[0]))

    while len(new_args) < len(args):
        arg = args[len(new_args)]

        if arg.contents == "then" or arg.contents == "else":
            new_args.append(arg)
            continue
        if arg.contents == "elseif":
            new_args.append(arg)
            new_args.append(parser.parse_expression(args[len(new_args)]))
            continue

        arg = parser.parse_script(arg)
        new_args.append(arg)

    return new_args


def _interp_eval(args, parser):
    if len(args) < 2:
        raise CommandArgError(
            f"not enough args to 'interp eval': got {len(args)}, expected at least 2"
        )
    return args[0:1] + eval(args[1:], parser, "interp eval")


def _lmap(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/lmap.html
    if len(args) < 3:
        raise CommandArgError(
            f"not enough args to lmap: got {len(args)}, expected at least 3"
        )

    return args[:-1] + [parser.parse_script(args[-1])]


def _namespace_code(args, parser):
    # ref: https://www.tcl.tk/man/tcl8.4/TclCmd/namespace.html#M6
    # TODO: seems like a possible pattern is to execute things in these scripts
    # with additional args provided, so command-args checks within this might
    # actually be false positive. will keep as-is for now though.
    return [parser.parse_script(args[0])]


def _namespace_eval(args, parser):
    if len(args) < 2:
        raise CommandArgError(
            f"not enough args to 'namespace eval': got {len(args)}, expected at least 2"
        )
    return args[0:1] + eval(args[1:], parser, "namespace eval")


def _namespace_inscope(args, parser):
    # ref: https://www.tcl.tk/man/tcl8.4/TclCmd/namespace.html#M14
    raise CommandArgError(
        "'namespace inscope' is not meant to be called directly, consider using"
        " 'namespace code' or 'namespace eval' instead"
    )


def _package_ifneeded(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/package.html

    # TODO: implement

    # one issue with this one - it seems like calls to package ifneeded are
    # often generated by pkg_MkIndex and these calls won't lint clean. Probably
    # need a special case to ensure that these don't generate violations

    raise CommandArgError(
        "argument parsing for 'package ifneeded' not implemented, any script argument"
        " will not be checked for violations"
    )


def _proc(args, parser):
    if len(args) != 3:
        raise CommandArgError(f"wrong # of args to proc: got {len(args)}, expected 3")

    # Parse args as list, then iterate over each item to parse arg specifier lists and
    # do some validation. We don't store non-defaulted arguments as Lists so that they
    # don't get formatted inside braces.
    arg_list = parser.parse_list(args[1])
    for i, arg in enumerate(arg_list.children):
        if isinstance(arg, BareWord):
            continue

        arg_specifier = parser.parse_list(arg)
        arg_specifier_len = len(arg_specifier.children)

        if arg_specifier_len == 2:
            arg_list.children[i] = arg_specifier
        elif arg_specifier_len != 1:
            raise CommandArgError(
                f"too many fields in argument specifier: got {arg_specifier_len},"
                " expected no more than 2"
            )

    return args[0:1] + [arg_list, parser.parse_script(args[2])]


def _return(args, parser):
    args = list(args)
    while len(args) > 0:
        option = args.pop(0).contents

        try:
            if option == "-code":
                arg = args.pop(0)
                try:
                    _check_code(arg)
                except CommandArgError as e:
                    raise CommandArgError(f"invalid value for return -code: {e}")
            elif option == "-level":
                val = args.pop(0).contents

                if val is None:
                    continue

                try:
                    if int(val) >= 0:
                        continue
                except ValueError:
                    pass

                raise CommandArgError(
                    f"invalid value for return -level: got {val}, expected a"
                    " non-negative integer"
                )
            elif option in {"-errorcode", "-errorinfo", "-errorstack", "-options"}:
                args.pop(0)
            else:
                break
        except IndexError:
            raise CommandArgError(
                f"insufficient args to return: expected value after {option}"
            )

    if len(args) > 0:
        raise CommandArgError(
            "too many arguments to return: expected no more than 1 argument after"
            " explicit options. Provide -options argument if you intend to specify"
            " additional return options."
        )

    return None


def _switch(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/switch.html
    # This one's complicated...

    # TODO: better checking of malformed switch command

    arg_contents = [arg.contents for arg in args]
    arg_i = 0

    try:
        arg_i = arg_contents.index("--") + 1
    except ValueError:
        while True:
            contents = args[arg_i].contents
            if contents in {"-exact", "-glob", "-regexp", "-nocase"}:
                arg_i += 1
            elif contents in {"-matchvar", "-indexvar"}:
                arg_i += 2
            else:
                break

    # accounts for string to be matched
    arg_i += 1

    new_args = args[0:arg_i]

    # one argument left => form where patterns and bodies are in list
    last_arg_is_list = arg_i == len(args) - 1

    if last_arg_is_list:
        pattern_and_commands_list = parser.parse_list(args[arg_i])
        new_args.append(pattern_and_commands_list)
        pattern_and_commands = pattern_and_commands_list.children
    else:
        pattern_and_commands = args[arg_i:]

    if len(pattern_and_commands) % 2 != 0:
        raise CommandArgError("Expected even number of patterns and commands")

    parsed_patterns_and_commands = []
    for i, node in enumerate(pattern_and_commands):
        if i % 2 == 0:
            parsed_patterns_and_commands.append(node)
        else:
            parsed_patterns_and_commands.append(parser.parse_script(node))

    if last_arg_is_list:
        pattern_and_commands_list.children = parsed_patterns_and_commands
    else:
        new_args.extend(parsed_patterns_and_commands)

    return new_args


def _time(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/time.html
    if len(args) < 1:
        raise CommandArgError(
            f"not enough args to time: got {len(args)}, expected at least 1"
        )

    if len(args) > 2:
        raise CommandArgError(
            f"too many args to time: got {len(args)}, expected no more than 2"
        )

    if len(args) == 2:
        time = args[1].contents
        if time is not None:
            try:
                int(time)
            except ValueError:
                raise CommandArgError(
                    "invalid argument to time: expected integer for last argument"
                )

    return [parser.parse_script(args[0])] + args[1:]


def _timerate(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/timerate.html
    # timerate doesn't seem to be implemented in tclsh 8.6 for me - why?

    args = list(args)
    new_args = []

    while True:
        try:
            arg = args.pop(0)
        except IndexError:
            raise CommandArgError("invalid arguments to timerate: expected script body")

        if arg.contents in {"-direct", "-calibrate"}:
            new_args.append(arg)
        elif arg.contents in {"-overhead"}:
            new_args.append(arg)
            try:
                val = args.pop(0)
                if val.contents is not None:
                    float(val.contents)
            except (ValueError, IndexError, TypeError):
                raise CommandArgError(
                    "invalid argument to timerate: -overhead must be followed by a"
                    " double"
                )
            new_args.append(val)
        else:
            break

    new_args.append(parser.parse_script(arg))

    if len(args) > 2:
        raise CommandArgError(
            "too many arguments to timerate: expected no more than 2 arguments"
            " following script body"
        )

    try:
        [int(arg.contents) for arg in args]
    except ValueError:
        raise CommandArgError(
            "invalid argument to timerate: expected one or two integers following"
            " script body"
        )

    return new_args + args


def _try(args, parser):
    # ref: https://www.tcl.tk/man/tcl/TclCmd/try.html
    args = list(args)
    new_args = []

    while True:
        try:
            arg = args.pop(0)
        except IndexError:
            raise CommandArgError("invalid arguments to try: missing script body")
        new_args.append(parser.parse_script(arg))

        try:
            arg = args.pop(0)
        except IndexError:
            break

        new_args.append(arg)

        if arg.contents == "on":
            try:
                code = args.pop(0)
                try:
                    _check_code(code)
                except CommandArgError as e:
                    raise CommandArgError(
                        f"invalid code argument to 'on' handler in try: {e}"
                    )
                new_args.append(code)
                new_args.append(args.pop(0))
            except IndexError:
                raise CommandArgError(
                    "invalid arguments to try: expected 3 arguments after 'on' handler"
                )
        elif arg.contents == "trap":
            try:
                new_args.append(args.pop(0))
                new_args.append(args.pop(0))
            except IndexError:
                raise CommandArgError(
                    "invalid arguments to try: expected 3 arguments after 'trap'"
                    " handler"
                )
        elif arg.contents == "finally":
            continue
        else:
            raise CommandArgError(
                "invalid handler argument to try: expected one of 'on', 'trap', or"
                " 'finally'"
            )

    return new_args


def _while(args, parser):
    if len(args) != 2:
        raise CommandArgError(f"wrong # of args to while: got {len(args)}, expected 2")

    return [
        parser.parse_expression(args[0]),
        parser.parse_script(args[1]),
    ]


commands = {
    "after": subcommands(
        "after",
        {
            "cancel": _after_cancel,
            "idle": _after_idle,
            "info": check_count("after info", 0, 1),
        },
        default=_after,
    ),
    "append": check_count("append", 1, None),
    "apply": _apply,
    # TODO: check subcommands
    "array": check_count("array", None, None),
    "binary": subcommands(
        "binary",
        {
            "decode": check_count("binary decode", 2, None),
            "encode": check_count("binary encode", 2, None),
            "format": check_count("binary format", 1, None),
            "scan": check_count("binary scan", 2, None),
        },
    ),
    "break": check_count("break", 0, 0),
    "catch": _catch,
    "cd": check_count("cd", 0, 1),
    # TODO: check subcommands
    "chan": check_count("chan"),
    # TODO: check subcommands
    "clock": check_count("clock"),
    "close": check_count("close", 1, 2),
    "concat": check_count("concat"),
    "continue": check_count("continue", 0, 0),
    "coroutine": check_count("coroutine", 2, None),
    "dict": subcommands(
        "dict",
        {
            "append": check_count("dict append", 2, None),
            "create": check_count("dict create"),
            "exists": check_count("dict exists", 2, None),
            "filter": _dict_filter,
            "for": _dict_map_for("dict for"),
            "get": check_count("dict get", 1, None),
            "incr": check_count("dict incr", 2, 3),
            "info": check_count("dict info", 1, 1),
            "keys": check_count("dict keys", 1, 2),
            "lappend": check_count("dict lappend", 2, None),
            "map": _dict_map_for("dict map"),
            "merge": check_count("dict merge"),
            "remove": check_count("dict remove", 1, None),
            "replace": check_count("dict replace", 1, None),
            "set": check_count("dict set", 3, None),
            "size": check_count("dict size", 1, 1),
            "unset": check_count("dict unset", 2, None),
            "update": _dict_update,
            "values": check_count("dict values", 1, 2),
            "with": _dict_with,
        },
    ),
    "encoding": subcommands(
        "encoding",
        {
            "convertfrom": check_count("encoding convertfrom", 1, 2),
            "convertto": check_count("encoding convertto", 1, 2),
            "dirs": check_count("encoding dirs", 0, 1),
            "names": check_count("encoding names", 0, 0),
            "system": check_count("encoding system", 0, 1),
        },
    ),
    "eof": check_count("eof", 1, 1),
    "error": check_count("error", 1, 3),
    "eval": _eval,
    "exec": check_count("exec", 1, None),
    "exit": check_count("exit", 0, 1),
    "expr": _expr,
    "fblocked": check_count("fblocked", 1, 1),
    "fconfigure": check_count("fconfigure", 1, None),
    "fcopy": check_count("fcopy", 2, 6),
    # TODO: check subcommands
    "file": check_count("file", 1, None),
    "fileevent": _fileevent,
    "flush": check_count("flush", 1, 1),
    "for": _for,
    "foreach": _foreach,
    "format": check_count("format", 1, None),
    "gets": check_count("gets", 1, 2),
    "glob": check_count("glob"),
    "global": check_count("global"),
    "history": check_count("history"),
    "if": _if,
    "incr": check_count("incr", 1, 2),
    # TODO: check subcommands
    "info": check_count("info", 1, None),
    # TODO: check other subcommands
    "interp": subcommands(
        "interp",
        {
            "eval": _interp_eval,
        },
        default=check_count("interp", 1, None),
    ),
    "join": check_count("join", 1, 2),
    "lappend": check_count("lappend", 1, None),
    "lassign": check_count("lassign", 1, None),
    "lindex": check_count("lindex", 1, None),
    "linsert": check_count("linsert", 2, None),
    "list": check_count("list", 0, None),
    "llength": check_count("llength", 1, 1),
    "lrepeat": check_count("lrepeat", 1, None),
    "lreplace": check_count("lreplace", 3, None),
    "lreverse": check_count("lreverse", 1, 1),
    "lset": check_count("lset", 2, None),
    "lsort": check_count("lsort", 1, None),
    "lmap": _lmap,
    "load": check_count("load", 1, 6),
    "lrange": check_count("lrange", 3, 3),
    "lsearch": check_count("lsearch", 2, None),
    "memory": subcommands(
        "memory",
        {
            "active": check_count("memory active", 1, 1),
            "break_on_malloc": check_count("memory break_on_malloc", 1, 1),
            "info": check_count("memory info", 0, 0),
            # just on or off
            "init": check_count("memory init", 1, 1),
            "objs": check_count("memory objs", 1, 1),
            "onexit": check_count("memory onexit", 1, 1),
            "tag": check_count("memory tag", 1, 1),
            # just on or off
            "trace": check_count("memory trace", 1, 1),
            "trace_on_at_malloc": check_count("memory trace_on_at_malloc", 1, 1),
            # just on or off
            "validate": check_count("memory validate", 1, 1),
        },
    ),
    "namespace": subcommands(
        "namespace",
        {
            "children": check_count("namespace children", 0, 2),
            "code": _namespace_code,
            "current": check_count("namespace current", 0, 0),
            "delete": None,
            "eval": _namespace_eval,
            "exists": check_count("namespace exists", 1, 1),
            "export": None,
            "forget": None,
            "import": None,
            "inscope": _namespace_inscope,
            "origin": check_count("namespace origin", 1, 1),
            "parent": check_count("namespace parent", 0, 1),
            "qualifiers": check_count("namespace qualifiers", 1, 1),
            "tail": check_count("namespace tail", 1, 1),
            "which": check_count("namespace which", 1, 2),
            "ensemble": subcommands(
                "namespace ensemble",
                {
                    "create": None,
                    "configure": check_count("namespace ensemble configure", 1, None),
                    "exists": check_count("namespace ensemble exists", 1, 1),
                },
            ),
        },
    ),
    "open": check_count("open", 1, 3),
    "package": subcommands(
        "package",
        {
            "forget": None,
            "ifneeded": _package_ifneeded,
            "names": check_count("package names", 0, 0),
            "present": check_count("package present", 0, None),
            "provide": check_count("package provide", 1, 2),
            "require": check_count("package require", 1, None),
            "unknown": check_count("package unknown", 1, None),
            "vcompare": check_count("package vcompare", 2, 2),
            "versions": check_count("package versions", 1, 1),
            "vsatisfies": check_count("package vsatisfies", 2, None),
            "prefer": check_count("package prefer", 1, 1),
        },
    ),
    "pid": check_count("pid", 0, 1),
    "pkg::create": check_count("pkg::create", 2, None),
    "pkg_mkIndex": check_count("pkg_mkIndex", 1, None),
    "proc": _proc,
    "puts": check_count("puts", 1, 3),
    "pwd": check_count("pwd", 0, 0),
    "read": check_count("read", 1, 2),
    "regexp": check_count("regexp", 2, None),
    "regsub": check_count("regsub", 3, None),
    "rename": check_count("rename", 2, 2),
    "return": _return,
    # TODO: check subcommands
    "safe": check_count("safe", 1, None),
    "scan": check_count("scan", 2, None),
    "seek": check_count("seek", 2, 3),
    "set": check_count("set", 1, 2),
    "socket": check_count("socket", 2, None),
    "source": check_count("source", 1, 3),
    "split": check_count("split", 1, 2),
    # TODO: check subcommands
    "string": check_count("string", 2, None),
    "subst": check_count("subst", 1, 4),
    "switch": _switch,
    "tailcall": check_count("tailcall", 1, None),
    "tcl::prefix": subcommands(
        "tcl::prefix",
        {
            "all": check_count("tcl::prefix all", 2, 2),
            "longest": check_count("tcl::prefix longest", 2, 2),
            "match": check_count("tcl::prefix match", 2, None),
        },
    ),
    "tell": check_count("tell", 1, 1),
    "throw": check_count("throw", 2, 2),
    "time": _time,
    "timerate": _timerate,
    "tcl::tm::path": subcommands(
        "tcl::tm::path",
        {
            "add": check_count("tcl::tm::path add"),
            "remove": check_count("tcl::tm::path remove"),
            "list": check_count("tcl::tm::path list", 0, 0),
        },
    ),
    "tcl::tm::roots": check_count("tcl::tm::roots"),
    # TODO: check subcommands
    "trace": check_count("trace", 2, None),
    "try": _try,
    "unload": check_count("unload", 1, 6),
    "unset": check_count("unset"),
    "update": check_count("update", 0, 1),
    "uplevel": check_count("uplevel", 1, None),
    "upvar": check_count("upvar", 2, None),
    "variable": check_count("variable", 1, None),
    "vwait": check_count("vwait", 1, 1),
    "while": _while,
    "yield": check_count("yield", 0, 1),
    "yieldto": check_count("yieldto", 2, None),
    # TODO: check subcommands
    "zlib": check_count("zlib", 3, None),
}



"""CLI utility for formatting Tcl code."""

import argparse
import pathlib
import sys

from tclint.cli.utils import resolve_sources, register_codec_warning
from tclint.config import (
    get_config,
    setup_tclfmt_config_cli_args,
    Config,
    ConfigError,
    RunConfig,
)
from tclint.parser import Parser, TclSyntaxError
from tclint.format import Formatter, FormatterOpts

try:
    from tclint._version import __version__  # type: ignore
except ModuleNotFoundError:
    __version__ = "(unknown version)"

# exit code flags
EXIT_OK = 0
EXIT_FORMAT_VIOLATIONS = 1
EXIT_SYNTAX_ERROR = 2
EXIT_INPUT_ERROR = 4


def format(script: str, config: Config, debug=False) -> str:
    plugins = [config.commands] if config.commands is not None else []
    parser = Parser(debug=debug, command_plugins=plugins)

    formatter = Formatter(
        FormatterOpts(
            indent=config.get_indent(),
            spaces_in_braces=config.style_spaces_in_braces,
            max_blank_lines=config.style_max_blank_lines,
            indent_namespace_eval=config.style_indent_namespace_eval,
        )
    )
    return formatter.format_top(script, parser)


def check(path: pathlib.Path, script: str, formatted: str):
    parser = Parser()
    original_tree = parser.parse(script)
    formatted_tree = parser.parse(formatted)
    if original_tree != formatted_tree:
        print(f"Warning: {path} syntax trees don't match", file=sys.stderr)
        print("\n".join(original_tree.diff(formatted_tree)), file=sys.stderr)


def main():
    parser = argparse.ArgumentParser("tclfmt")
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "source",
        nargs="+",
        help=(
            "files to format. By default, prints formatted files to stdout. Provide '-'"
            " to read from stdin"
        ),
        type=pathlib.Path,
    )

    mode_group = parser.add_argument_group("mode")
    mode_mutex = mode_group.add_mutually_exclusive_group(required=False)
    mode_mutex.add_argument(
        "--in-place", help="update files that require formatting", action="store_true"
    )
    mode_mutex.add_argument(
        "--check",
        help="list files that require formatting and set the exit code",
        action="store_true",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="count",
        default=0,
        help=(
            "display debug output. Provide additional times to increase the verbosity"
            " of output (e.g. -dd)"
        ),
    )
    parser.add_argument(
        "-c",
        "--config",
        help="path to config file",
        type=pathlib.Path,
        default=None,
        metavar="<path>",
    )
    setup_tclfmt_config_cli_args(parser)
    args = parser.parse_args()

    try:
        config = get_config(args.config, pathlib.Path.cwd())
    except ConfigError as e:
        print(f"Invalid config file: {e}")
        return EXIT_INPUT_ERROR

    if config is None:
        config = RunConfig()

    config.apply_cli_args(args)

    try:
        # TODO: we should eventually allow tclfmt to find a config by walking up
        # directories, at which point exclude_root should be the parent dir of
        # the config file, unless -c is used (eslint rules)
        exclude_root = pathlib.Path.cwd()
        sources = resolve_sources(
            args.source,
            exclude_patterns=config.exclude,
            exclude_root=exclude_root,
            extensions=config.extensions,
        )
    except FileNotFoundError as e:
        print(f"Invalid path provided: {e}")
        return EXIT_INPUT_ERROR

    retcode = EXIT_OK

    register_codec_warning("replace_with_warning")

    reformat_count = 0
    for path in sources:
        if path is None:
            script = sys.stdin.read()
            out_prefix = "(stdin)"
        else:
            with open(path, "r", errors="replace_with_warning") as f:
                script = f.read()
            out_prefix = str(path)

        try:
            formatted = format(
                script, config.get_for_path(path), debug=(args.debug > 1)
            )
            if args.in_place and path:
                with open(path, "w") as f:
                    f.write(formatted)
            elif args.check:
                if script != formatted:
                    print(f"{out_prefix}: needs reformatting")
                    retcode |= EXIT_FORMAT_VIOLATIONS
                    reformat_count += 1
            else:
                if args.in_place:
                    print("Warning: --in-place option ignored when reading from stdin")
                print(formatted, end="")

            if args.debug > 0:
                check(path, script, formatted)
        except TclSyntaxError as e:
            line, col = e.pos
            print(f"{out_prefix}:{line}:{col}: syntax error: {e}", file=sys.stderr)
            retcode |= EXIT_SYNTAX_ERROR
            continue

    if args.check:
        messages = []
        if reformat_count == 0:
            messages.append("Formatting clean!")
        elif reformat_count == 1:
            messages.append("1 file needs reformatting.")
        else:
            messages.append(f"{reformat_count} files need reformatting.")
        messages.append(
            f"Checked {len(sources)} file{'s' if len(sources) != 1 else ''}."
        )
        print(" ".join(messages))

    return retcode


if __name__ == "__main__":
    sys.exit(main())



import pytest
import jsonschema

from tclint.commands.schema import schema


def test_valid_command_spec():
    command_spec = {
        "plugin": "test_plugin",
        "spec": {
            "command1": {
                "": {"min": 0, "max": None},
                "-switch1": {"required": False, "value": True, "repeated": False},
                "-switch2": {"required": True, "value": False, "repeated": True},
            }
        },
    }

    jsonschema.validate(instance=command_spec, schema=schema)


def test_missing_required_field():
    invalid_command_spec = {
        "plugin": "test_plugin",
        "spec": {
            "command1": {
                "": {"min": 0, "max": None},
                "-switch3": {"value": True, "repeated": False},
            }
        },
    }

    with pytest.raises(jsonschema.exceptions.ValidationError):
        jsonschema.validate(instance=invalid_command_spec, schema=schema)


def test_invalid_minmax_values():
    invalid_minmax_spec = {
        "plugin": "test_plugin",
        "spec": {"command1": {"": {"min": "0", "max": None}}},  # Should be integer
    }

    with pytest.raises(jsonschema.exceptions.ValidationError):
        jsonschema.validate(instance=invalid_minmax_spec, schema=schema)


def test_extra_property():
    invalid_spec = {
        "plugin": "test_plugin",
        "spec": {
            "command1": {
                "": {
                    "min": 0,
                    "max": None,
                    "extra_field": "not allowed",  # Extra property not in schema
                }
            }
        },
    }

    with pytest.raises(jsonschema.exceptions.ValidationError):
        jsonschema.validate(instance=invalid_spec, schema=schema)



from pathlib import Path
import shutil
import sys

from lsprotocol import types as lsp
import pytest
import pytest_lsp

MY_DIR = Path(__file__).parent.resolve()
LSP_BIN = MY_DIR / ".." / "src" / "tclint" / "cli" / "tclsp.py"


def get_capabilities(client: str) -> lsp.ClientCapabilities:
    capabilities = pytest_lsp.client_capabilities(client)
    # Need to nullify these capabilities, since pytest_lsp doesn't support them
    try:
        capabilities.workspace.diagnostics.refresh_support = False
    except AttributeError:
        pass
    try:
        capabilities.workspace.did_change_watched_files.dynamic_registration = False
    except AttributeError:
        pass
    return capabilities


@pytest_lsp.fixture(
    config=pytest_lsp.ClientServerConfig(server_command=[sys.executable, str(LSP_BIN)])
)
async def client(lsp_client: pytest_lsp.LanguageClient, request):
    yield

    # Teardown
    await lsp_client.shutdown_session()


@pytest.mark.asyncio
async def test_diagnostics(client: pytest_lsp.LanguageClient):
    """Basic diagnostics test."""
    params = lsp.InitializeParams(capabilities=get_capabilities("visual-studio-code"))
    await client.initialize_session(params)

    document = MY_DIR / "data" / "dirty.tcl"
    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=f"file://{document}")
        )
    )
    assert results is not None
    assert len(results.items) == 2
    items = sorted(
        results.items, key=lambda r: (r.range.start.line, r.range.start.character)
    )

    assert items[0].message == "unnecessary command substitution within expression"
    assert items[0].code == "redundant-expr"
    assert items[0].range.start.line == 5
    assert items[0].range.start.character == 10
    assert items[0].range.end.line == 5
    assert items[0].range.end.character == 23

    assert (
        items[1].message == "expression with substitutions should be enclosed by braces"
    )
    assert items[1].code == "unbraced-expr"
    assert items[1].range.start.line == 5
    assert items[1].range.start.character == 16
    assert items[1].range.end.line == 5
    assert items[1].range.end.character == 22


@pytest.mark.asyncio
async def test_format(client: pytest_lsp.LanguageClient, tmp_path):
    """Formatting test."""
    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        workspace_folders=[
            lsp.WorkspaceFolder(uri=tmp_path.as_uri(), name=tmp_path.name)
        ],
    )
    await client.initialize_session(params)

    document = tmp_path / "source.tcl"

    INDENT_TEST = """if {{1}} {{
{}puts "hello"
}}
"""
    for source, expected, end in (
        ("foo", "foo\n", (0, 3)),
        ("foo\n\n\n\n\nbar\n", "foo\n\n\nbar\n", (6, 0)),
        (INDENT_TEST.format(""), INDENT_TEST.format("   "), (3, 0)),
    ):
        with open(document, "w") as f:
            f.write(source)

        result = await client.text_document_formatting_async(
            params=lsp.DocumentFormattingParams(
                text_document=lsp.TextDocumentIdentifier(uri=document.as_uri()),
                options=lsp.FormattingOptions(tab_size=3, insert_spaces=True),
            ),
        )
        assert len(result) == 1
        result = result[0]

        assert result.new_text == expected
        assert result.range.start.line == 0
        assert result.range.start.character == 0
        assert result.range.end.line == end[0]
        assert result.range.end.character == end[1]

    # Check that tclint config overrides client config
    config = tmp_path / "tclint.toml"
    with open(config, "w") as f:
        f.write("""[style]
indent = 'tab'""")

    client.workspace_did_change_watched_files(
        # `changes` isn't used by our server impl, so we can cheat and keep this empty
        params=lsp.DidChangeWatchedFilesParams(changes=[])
    )

    result = await client.text_document_formatting_async(
        params=lsp.DocumentFormattingParams(
            text_document=lsp.TextDocumentIdentifier(uri=document.as_uri()),
            options=lsp.FormattingOptions(tab_size=3, insert_spaces=True),
        ),
    )
    assert len(result) == 1
    result = result[0]

    assert result.new_text == INDENT_TEST.format("\t")
    assert result.range.start.line == 0
    assert result.range.start.character == 0
    assert result.range.end.line == 3
    assert result.range.end.character == 0


@pytest.mark.asyncio
async def test_config(client: pytest_lsp.LanguageClient, tmp_path_factory):
    """Tests configuration reading, specifically:
    1) That configs for multiple workspaces are found and applied correctly.
    2) That diagnostics are refreshed correctly when config is updated.

    Note proper use of the client capabilities we rely on for file watching/diagnostic
    refresh aren't captured by this test, the client is controlled manually.
    """
    # Set up workspaces
    document = MY_DIR / "data" / "dirty.tcl"

    ws_foo = tmp_path_factory.mktemp("foo")
    with open(ws_foo / "tclint.toml", "w") as f:
        f.write("ignore = ['redundant-expr']")
    doc_foo = Path(shutil.copy(document, ws_foo))

    ws_bar = tmp_path_factory.mktemp("bar")
    with open(ws_bar / "tclint.toml", "w") as f:
        f.write("ignore = ['unbraced-expr']")
    doc_bar = Path(shutil.copy(document, ws_bar))

    # Initialize client
    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        workspace_folders=[
            lsp.WorkspaceFolder(uri=ws_foo.as_uri(), name=ws_foo.name),
            lsp.WorkspaceFolder(uri=ws_bar.as_uri(), name=ws_bar.name),
        ],
    )
    await client.initialize_session(params)

    # Check diagnostics for workspace "foo"
    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=doc_foo.as_uri())
        )
    )
    assert results is not None
    assert len(results.items) == 1
    assert results.items[0].code == "unbraced-expr"

    # Check diagnostics for workspace "bar"
    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=doc_bar.as_uri())
        )
    )
    assert results is not None
    assert len(results.items) == 1
    assert results.items[0].code == "redundant-expr"

    # Make sure diagnostics refresh when config file is updated
    with open(ws_foo / "tclint.toml", "w") as f:
        f.write("ignore = ['redundant-expr', 'unbraced-expr']")
    client.workspace_did_change_watched_files(
        # `changes` isn't used by our server impl, so we can cheat and keep this empty
        params=lsp.DidChangeWatchedFilesParams(changes=[])
    )

    results = await client.text_document_diagnostic_async(
        params=lsp.DocumentDiagnosticParams(
            text_document=lsp.TextDocumentIdentifier(uri=doc_foo.as_uri())
        )
    )
    assert results.items == []


@pytest.mark.asyncio
async def test_invalid_config(client: pytest_lsp.LanguageClient, tmp_path_factory):
    # Set up workspace
    ws = tmp_path_factory.mktemp("ws")
    config = ws / "tclint.toml"
    # Invalid config
    with open(config, "w") as f:
        f.write("asdf")

    # Initialize client
    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        workspace_folders=[
            lsp.WorkspaceFolder(uri=ws.as_uri(), name=ws.name),
        ],
    )
    await client.initialize_session(params)
    await client.wait_for_notification(lsp.WINDOW_SHOW_MESSAGE)
    assert len(client.messages) == 1
    assert (
        client.messages[0].message
        == f"Error loading config file: {config}: Expected '=' after a key in a"
        " key/value pair (at end of document)"
    )


@pytest.mark.asyncio
async def test_nested_configs(client: pytest_lsp.LanguageClient, tmp_path_factory):
    # Set up workspaces
    parent_ws = tmp_path_factory.mktemp("ws")
    with open(parent_ws / "tclint.toml", "w") as f:
        f.write("")
    child_ws = parent_ws / "child"
    child_ws.mkdir()
    with open(child_ws / "tclint.toml", "w") as f:
        f.write("")

    # Initialize client
    params = lsp.InitializeParams(
        capabilities=get_capabilities("visual-studio-code"),
        workspace_folders=[
            lsp.WorkspaceFolder(uri=parent_ws.as_uri(), name=parent_ws.name),
            lsp.WorkspaceFolder(uri=child_ws.as_uri(), name=child_ws.name),
        ],
    )
    await client.initialize_session(params)
    await client.wait_for_notification(lsp.WINDOW_SHOW_MESSAGE)
    assert len(client.messages) == 1
    assert (
        client.messages[0].message
        == f"Warning: found configs in overlapping workspaces: {child_ws},"
        f" {parent_ws}. It's undefined which will apply."
    )



"""JSON Schema defining command specs."""

_command_schema = {
    "type": "object",
    "properties": {
        "": {
            "type": "object",
            "properties": {
                "min": {
                    "type": "integer",
                    "description": "Minimum number of arguments",
                },
                "max": {
                    "type": ["integer", "null"],
                    "description": "Maximum number of arguments. Null means unlimited.",
                },
            },
            "additionalProperties": False,
            "required": ["min", "max"],
        }
    },
    "patternProperties": {
        "^.+": {
            "type": "object",
            "properties": {
                "required": {
                    "type": "boolean",
                    "description": "Whether this switch is required",
                },
                "value": {
                    "type": "boolean",
                    "description": "Whether this switch takes a value",
                },
                "repeated": {
                    "type": "boolean",
                    "description": "Whether this switch can be repeated",
                },
            },
            "additionalProperties": False,
            "required": ["required", "value", "repeated"],
        }
    },
}

schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "title": "Command Spec",
    "description": "Describes a set of commands and their arguments.",
    "type": "object",
    "properties": {
        "plugin": {
            "type": "string",
            "description": "The name of the plugin associated with this command spec.",
        },
        "spec": {"type": "object", "patternProperties": {"^.+": _command_schema}},
    },
    "additionalProperties": False,
}



import subprocess
import re
from typing import Optional, List, Dict

from tclint.plugins.openroad.help_parser import spec_from_help_entry
from tclint.commands.utils import CommandArgError, check_count
from tclint.syntax_tree import BareWord


def check_arg_spec(command, arg_spec):
    # TODO check required arguments
    def check(args, parser):
        args_allowed = set(arg_spec.keys())
        positional_args = []

        args = list(args)
        while len(args) > 0:
            arg = args.pop(0)

            # To facilitate better error messages, we expect that switches are always
            # specified as BareWords that start with "-" or ">". This lets us throw an
            # error when a switch-like thing doesn't match any supported arguments,
            # rather than counting it towards the positional arguments (which usually
            # ends up in a vague "too many arguments" error). To make tclint interpret a
            # switch-like word as a positional argument, users should wrap it in "", and
            # any switches should be BareWords.
            contents = arg.contents
            if not (
                isinstance(arg, BareWord) and contents and contents[0] in {"-", ">"}
            ):
                positional_args.append(arg)
                continue

            if contents in args_allowed:
                if arg_spec[contents]["value"]:
                    try:
                        args.pop(0)
                    except IndexError:
                        raise CommandArgError(
                            f"invalid arguments for {command}: expected value after"
                            f" {contents}"
                        )
                if not arg_spec[contents]["repeated"]:
                    args_allowed.remove(contents)
            elif contents in arg_spec:
                raise CommandArgError(f"duplicate argument for {command}: {contents}")
            else:
                prefix_matches = []
                for switch in arg_spec:
                    if switch.startswith(contents):
                        prefix_matches.append(switch)

                if len(prefix_matches) == 1:
                    raise CommandArgError(
                        f"shortened argument for {command}: expand {contents} to"
                        f" {prefix_matches[0]}"
                    )

                if len(prefix_matches) > 1:
                    raise CommandArgError(
                        f"ambiguous argument for {command}: {contents} could be any of"
                        f" {', '.join(prefix_matches)}"
                    )

                raise CommandArgError(
                    f"unrecognized argument for {command}: {contents}"
                )

        check = check_count(
            command,
            min=arg_spec[""]["min"],
            max=arg_spec[""]["max"],
            args_name="positional args",
        )
        check(positional_args, parser)

        return None

    return check


_patches = {
    # missing {} around -box metavars
    "detailed_route_debug": {
        "": {"min": 0, "max": 0},
        "-pa": {"required": False, "value": False, "repeated": False},
        "-ta": {"required": False, "value": False, "repeated": False},
        "-dr": {"required": False, "value": False, "repeated": False},
        "-maze": {"required": False, "value": False, "repeated": False},
        "-net": {"required": False, "value": True, "repeated": False},
        "-pin": {"required": False, "value": True, "repeated": False},
        "-box": {"required": False, "value": True, "repeated": False},
        "-dump_last_worker": {"required": False, "value": False, "repeated": False},
        "-iter": {"required": False, "value": True, "repeated": False},
        "-pa_markers": {"required": False, "value": False, "repeated": False},
        "-dump_dr": {"required": False, "value": False, "repeated": False},
        "-dump_dir": {"required": False, "value": True, "repeated": False},
        "-pa_edge": {"required": False, "value": False, "repeated": False},
        "-pa_commit": {"required": False, "value": False, "repeated": False},
        "-write_net_tracks": {"required": False, "value": False, "repeated": False},
    },
    # need to handle pair of switches in exclusive group
    "remove_from_physical_cluster": {
        "": {"min": 1, "max": 1},
        "-parent_module": {"required": False, "value": True, "repeated": False},
        "-modinst": {"required": False, "value": True, "repeated": False},
        "-inst": {"required": False, "value": True, "repeated": False},
        "-physical_cluster": {"required": False, "value": True, "repeated": False},
    },
    # need to handle parens around switches
    "set_bump": {
        "": {"min": 0, "max": 0},
        "-row": {"required": True, "value": True, "repeated": False},
        "-col": {"required": True, "value": True, "repeated": False},
        "-remove": {"required": False, "value": False, "repeated": False},
        "-power": {"required": False, "value": True, "repeated": False},
        "-ground": {"required": False, "value": True, "repeated": False},
        "-net": {"required": False, "value": True, "repeated": False},
    },
    "with_output_to_variable": {
        "": {"min": 1, "max": None},
    },
    # need to fix unclosed [ after -fields
    "report_path": {
        # pin ^|r|rise|v|f|fall
        "": {"min": 2, "max": 2},
        # exclusive
        "-min": {"required": False, "value": False, "repeated": False},
        "-max": {"required": False, "value": False, "repeated": False},
        "-format": {"required": False, "value": True, "repeated": False},
        "-fields": {"required": False, "value": True, "repeated": False},
        "-digits": {"required": False, "value": True, "repeated": False},
        "-no_line_splits": {"required": False, "value": False, "repeated": False},
    },
    # TODO: probably need to evaluate other SDC commands to see if they require
    # repeating values
    "set_clock_groups": {
        "": {"min": 0, "max": 0},
        "-name": {"required": False, "value": True, "repeated": False},
        "-logically_exclusive": {"required": False, "value": False, "repeated": False},
        "-physically_exclusive": {"required": False, "value": False, "repeated": False},
        "-asynchronous": {"required": False, "value": False, "repeated": False},
        "-allow_paths": {"required": False, "value": False, "repeated": False},
        "-comment": {"required": False, "value": True, "repeated": False},
        "-group": {"required": True, "value": True, "repeated": True},
    },
    # Need to fix help string to disambiguate that "-mirror" doesn't consume "name" as a
    # value
    "place_pad": {
        "": {"min": 1, "max": 1},
        "-master": {"required": False, "value": True, "repeated": False},
        "-row": {"required": True, "value": True, "repeated": False},
        "-location": {"required": True, "value": True, "repeated": False},
        "-mirror": {"required": True, "value": False, "repeated": False},
    },
    # help string is missing metavars for most switches that take a value
    "clock_tree_synthesis": {
        "": {"min": 0, "max": 0},
        "-root_buf": {"required": False, "value": True, "repeated": False},
        "-buf_list": {"required": False, "value": True, "repeated": False},
        "-wire_unit": {"required": False, "value": True, "repeated": False},
        "-clk_nets": {"required": False, "value": True, "repeated": False},
        "-sink_clustering_size": {"required": False, "value": True, "repeated": False},
        "-num_static_layers": {"required": False, "value": True, "repeated": False},
        "-sink_clustering_buffer": {
            "required": False,
            "value": True,
            "repeated": False,
        },
        "-distance_between_buffers": {
            "required": False,
            "value": True,
            "repeated": False,
        },
        "-branching_point_buffers_distance": {
            "required": False,
            "value": True,
            "repeated": False,
        },
        "-clustering_exponent": {"required": False, "value": True, "repeated": False},
        "-clustering_unbalance_ratio": {
            "required": False,
            "value": True,
            "repeated": False,
        },
        "-sink_clustering_max_diameter": {
            "required": False,
            "value": True,
            "repeated": False,
        },
        "-sink_clustering_levels": {
            "required": False,
            "value": True,
            "repeated": False,
        },
        "-tree_buf": {"required": False, "value": True, "repeated": False},
        "-sink_clustering_buffer_max_cap_derate": {
            "required": False,
            "value": True,
            "repeated": False,
        },
        "-delay_buffer_derate": {"required": False, "value": True, "repeated": False},
        "-post_cts_disable": {"required": False, "value": False, "repeated": False},
        "-sink_clustering_enable": {
            "required": False,
            "value": False,
            "repeated": False,
        },
        "-balance_levels": {"required": False, "value": False, "repeated": False},
        "-obstruction_aware": {"required": False, "value": False, "repeated": False},
        "-apply_ndr": {"required": False, "value": False, "repeated": False},
        "-dont_use_dummy_load": {"required": False, "value": False, "repeated": False},
    },
    # help string needs to present diode_cell as optional
    "repair_antennas": {
        "": {"min": 0, "max": 1},
        "-iterations": {"required": False, "value": True, "repeated": False},
        "-ratio_margin": {"required": False, "value": True, "repeated": False},
    },
    # Need a way to indicate repeated arguments are okay.
    # TODO: this was unnecessary prior to the merge of
    # https://github.com/The-OpenROAD-Project/OpenROAD/pull/5100, if we add a way to
    # apply these to specific versions, that should be enabled here.
    "remove_buffers": {"": {"min": 0, "max": None}},
}


def make_command_spec(exec: Optional[List[str]] = None) -> Dict:
    if exec is not None:
        cmd = exec
    else:
        cmd = ["openroad"]
    cmd += ["-no_init", "-no_splash"]
    input = "help; exit"

    try:
        p = subprocess.run(
            cmd,
            input=input.encode(),
            capture_output=True,
        )
    except FileNotFoundError:
        raise Exception("openroad not installed")

    if p.returncode != 0:
        raise Exception(
            f"openroad exited with non-zero status. stderr: {p.stderr.decode()}"
        )

    help = p.stdout.decode()

    # Preprocessing help output to make it parseable
    # - Strip out comments
    help = re.sub(r"#.*$", "", help, flags=re.MULTILINE)
    # - Remove blank lines
    help = re.sub(r"^\n", "", help, flags=re.MULTILINE)
    # - Collapse each command onto one line
    help = re.sub(r"\n +", " ", help)

    command_spec = {}

    for line in help.split("\n"):
        if not line or line.startswith("openroad>"):
            # skip empty line or prompt line
            continue

        command_name = line.split()[0]
        if command_name in _patches:
            command_spec[command_name] = _patches[command_name]
        else:
            try:
                command_spec.update(spec_from_help_entry(line))
            except Exception:
                print(
                    f"Warning: failed to parse help string for {command_name}, this"
                    " command will not be documented"
                )

    return command_spec


def commands_from_spec(command_spec):
    return {
        command: check_arg_spec(command, arg_spec)
        for command, arg_spec in command_spec.items()
    }



import dataclasses
from typing import List, Tuple, Union
import sys

from tclint.syntax_tree import (
    Node,
    Script,
    Command,
    Comment,
    CommandSub,
    BareWord,
    QuotedWord,
    BracedWord,
    CompoundBareWord,
    VarSub,
    ArgExpansion,
    Expression,
    BracedExpression,
    ParenExpression,
    UnaryOp,
    BinaryOp,
    TernaryOp,
    Function,
)
from tclint.parser import Parser
from tclint.syntax_tree import List as ListNode


@dataclasses.dataclass
class LiteralBlock:
    block: List[str]
    pos: Tuple[int, int]
    end_pos: Tuple[int, int]


@dataclasses.dataclass
class FormatterOpts:
    indent: str
    spaces_in_braces: bool
    max_blank_lines: int
    indent_namespace_eval: bool


class Formatter:
    def __init__(self, opts: FormatterOpts):
        self.opts = opts

    def _indent(self, lines: List[str], indent: str) -> List[str]:
        indented = []
        for line in lines:
            if line == "":
                indented.append("")
            else:
                indented.append(indent + line)

        return indented

    def _brace(self, lines: List[str]) -> List[str]:
        spaces_in_braces = " " if self.opts.spaces_in_braces else ""
        if lines == [""]:
            return ["{" + spaces_in_braces + "}"]

        braced_lines = lines[:]
        braced_lines[0] = "{" + spaces_in_braces + lines[0]
        braced_lines[-1] += spaces_in_braces + "}"
        return braced_lines

    def format(self, *nodes: Union[Node, LiteralBlock]) -> List[str]:
        formatted = []
        for node in nodes:
            if isinstance(node, Script):
                formatted += self.format_script(node)
            elif isinstance(node, Command):
                formatted += self.format_command(node)
            elif isinstance(node, Comment):
                formatted += self.format_comment(node)
            elif isinstance(node, CommandSub):
                formatted += self.format_command_sub(node)
            elif isinstance(node, BareWord):
                formatted += self.format_bare_word(node)
            elif isinstance(node, QuotedWord):
                formatted += self.format_quoted_word(node)
            elif isinstance(node, BracedWord):
                formatted += self.format_braced_word(node)
            elif isinstance(node, CompoundBareWord):
                formatted += self.format_compound_bare_word(node)
            elif isinstance(node, VarSub):
                formatted += self.format_var_sub(node)
            elif isinstance(node, ArgExpansion):
                formatted += self.format_arg_expansion(node)
            elif isinstance(node, ListNode):
                formatted += self.format_list(node)
            elif isinstance(node, Expression):
                formatted += self.format_expression(node)
            elif isinstance(node, BracedExpression):
                formatted += self.format_braced_expression(node)
            elif isinstance(node, ParenExpression):
                formatted += self.format_paren_expression(node)
            elif isinstance(node, UnaryOp):
                formatted += self.format_unary_op(node)
            elif isinstance(node, BinaryOp):
                formatted += self.format_binary_op(node)
            elif isinstance(node, TernaryOp):
                formatted += self.format_ternary_op(node)
            elif isinstance(node, Function):
                formatted += self.format_function(node)
            elif isinstance(node, LiteralBlock):
                formatted += node.block
            else:
                assert False, f"unrecognized node: {type(node)}"

        return formatted

    def format_top(self, script: str, parser: Parser) -> str:
        tree = parser.parse(script)
        self.script = script.split("\n")
        return "\n".join(self.format_script_contents(tree)) + "\n"

    def format_script_contents(self, script: Union[Script, CommandSub]) -> List[str]:
        to_format = []
        skip_formatting_start = None
        for child in script.children:
            if skip_formatting_start is None:
                to_format.append(child)

            if isinstance(child, Comment):
                if child.value.strip() == "tclfmt-disable":
                    if skip_formatting_start is not None:
                        print(
                            "Warning: encountered 'tclint-disable' while formatting is"
                            " already disabled, ignoring...",
                            file=sys.stderr,
                        )
                    else:
                        skip_formatting_start = child.pos[0]
                elif child.value.strip() == "tclfmt-enable":
                    if skip_formatting_start is None:
                        print(
                            "Warning: encountered 'tclint-enable' while formatting is"
                            " already disabled, ignoring...",
                            file=sys.stderr,
                        )
                    else:
                        skip_formatting_end = child.pos[0]
                        block = self.script[skip_formatting_start:skip_formatting_end]
                        to_format.append(
                            LiteralBlock(
                                block,
                                pos=(skip_formatting_start + 1, 1),
                                end_pos=(skip_formatting_end, 1),
                            )
                        )
                        skip_formatting_start = None

        if skip_formatting_start is not None:
            print("Warning: missing 'tclint-enable'", file=sys.stderr)
            to_format.append(
                LiteralBlock(
                    self.script[skip_formatting_start:],
                    pos=(skip_formatting_start + 1, 1),
                    end_pos=script.end_pos,
                )
            )

        formatted = [""]
        last_line = None
        for child in to_format:
            if last_line is not None:
                if last_line == child.pos[0]:
                    if isinstance(child, Comment):
                        formatted[-1] += " ;"
                    else:
                        formatted[-1] += "; "
                else:
                    newlines = child.pos[0] - last_line
                    newlines = min(newlines, self.opts.max_blank_lines + 1)
                    formatted.extend([""] * newlines)
            last_line = child.end_pos[0]

            lines = self.format(child)
            formatted[-1] += lines[0]
            formatted.extend(lines[1:])

        return formatted

    def format_script(self, script: Script, should_indent=True) -> List[str]:
        lines = self.format_script_contents(script)
        if script.pos[0] == script.end_pos[0]:
            return self._brace(lines)

        # Usually, we enforce that multi-line scripts start on a new line after the open
        # brace. However, if a comment was originally on the same line as the open brace
        # we preserve it, since it's probably meant to be associated with this line
        # (e.g. a tclint-disable-line).
        open_brace = "{"
        if (
            len(script.children) > 0
            and isinstance(script.children[0], Comment)
            and script.pos[0] == script.children[0].pos[0]
        ):
            open_brace += " " + lines[0]
            lines = lines[1:]

        if should_indent:
            return [open_brace] + self._indent(lines, self.opts.indent) + ["}"]
        else:
            return [open_brace] + lines + ["}"]

    def format_command(self, command) -> List[str]:
        # TODO: enforce in type
        assert len(command.children) > 0

        is_namespace_eval = (
            command.routine == "namespace"
            and len(command.args) > 0
            and command.args[0].contents == "eval"
        )
        should_indent = not is_namespace_eval or self.opts.indent_namespace_eval

        hanging_indent = False
        formatted = self.format(command.children[0])
        last_line = command.children[0].end_pos[0]
        for child in command.children[1:]:
            if isinstance(child, Script):
                child_lines = self.format_script(child, should_indent=should_indent)
            else:
                child_lines = self.format(child)

            if last_line == child.pos[0]:
                formatted[-1] += " "
                formatted[-1] += child_lines[0]
            else:
                formatted[-1] += " \\"
                formatted.append(self.opts.indent + child_lines[0])
                hanging_indent = True

            if hanging_indent:
                formatted.extend(self._indent(child_lines[1:], self.opts.indent))
            else:
                formatted.extend(child_lines[1:])

            last_line = child.end_pos[0]

        return formatted

    def format_comment(self, comment: Comment) -> List[str]:
        return [f"#{comment.value}"]

    def format_command_sub(self, command_sub):
        if len(command_sub.children) == 0:
            return ["[]"]

        formatted = []
        contents = self.format_script_contents(command_sub)
        if len(command_sub.children) > 1 and len(contents) > 1:
            formatted.append("[")
            formatted.extend(self._indent(contents, self.opts.indent))
            formatted.append("]")
        else:
            formatted.append("[" + contents[0])
            formatted.extend(contents[1:])
            formatted[-1] += "]"

        return formatted

    def format_bare_word(self, word) -> List[str]:
        # Property enforced by parser
        assert word.contents is not None
        return [word.contents]

    def format_quoted_word(self, word) -> List[str]:
        if word.contents is not None:
            return [f'"{word.contents}"']

        formatted = ""
        for child in word.children:
            formatted += "\n".join(self.format(child))

        return [f'"{formatted}"']

    def format_braced_word(self, word) -> List[str]:
        assert word.contents is not None
        return [f"{{{word.contents}}}"]

    def format_compound_bare_word(self, word) -> List[str]:
        formatted = [""]
        for child in word.children:
            child_lines = self.format(child)
            formatted[-1] += child_lines[0]
            formatted.extend(child_lines[1:])

        return formatted

    def format_var_sub(self, varsub) -> List[str]:
        # We might be able to make the formatter infer whether braces are required, and
        # remove them from the syntax tree. For now it's easier to just mimic the
        # original format.
        if varsub.braced:
            formatted = [f"${{{varsub.value}}}"]
        else:
            formatted = [f"${varsub.value}"]

        if varsub.children:
            # We just concatenate everything as is, since changes in whitespace are
            # semantically meaningful in this context. Any newlines are captured by
            # BareWords.
            formatted[-1] += "("
            for child in varsub.children:
                child_lines = self.format(child)
                formatted[-1] += child_lines[0]
                formatted.extend(child_lines[1:])
            formatted[-1] += ")"

        return formatted

    def format_arg_expansion(self, arg_expansion) -> List[str]:
        # TODO: enforce in type?
        assert len(arg_expansion.children) == 1
        lines = self.format(arg_expansion.children[0])
        lines[0] = "{*}" + lines[0]

        return lines

    def format_list(self, list_node) -> List[str]:
        # Similar to Script, but the contents are a bit more straightforward.
        contents = [""]
        last_line = None
        for child in list_node.children:
            if last_line is not None:
                if last_line == child.pos[0]:
                    contents[-1] += " "
                else:
                    newlines = child.pos[0] - last_line
                    newlines = min(newlines, 3)
                    contents.extend([""] * newlines)

            lines = self.format(child)
            contents[-1] += lines[0]
            contents.extend(lines[1:])

            last_line = child.end_pos[0]

        if list_node.pos[0] == list_node.end_pos[0]:
            return self._brace(contents)

        return ["{"] + self._indent(contents, self.opts.indent) + ["}"]

    def format_expression(self, expr) -> List[str]:
        formatted = [""]
        for child in expr.children:
            lines = self.format(child)
            formatted[-1] += lines[0]
            for line in lines[1:]:
                formatted[-1] += " \\"
                formatted += self._indent([line], self.opts.indent)

        # Trick: we know there are quotes around the expression if the start of the
        # expression is a different column than its first child.
        quoted = expr.pos[1] != expr.children[0].pos[1]
        if quoted:
            formatted[0] = '"' + formatted[0]
            formatted[-1] += '"'

        return formatted

    def format_braced_expression(self, expr) -> List[str]:
        formatted = [""]
        for child in expr.children:
            lines = self.format(child)
            formatted[-1] += lines[0]
            formatted.extend(lines[1:])

        if expr.pos[0] == expr.end_pos[0]:
            return self._brace(formatted)

        return ["{"] + self._indent(formatted, self.opts.indent) + ["}"]

    def format_paren_expression(self, expr) -> List[str]:
        # TODO: enforce in type?
        assert len(expr.children) == 1
        body = expr.children[0]

        formatted = ["("]
        lines = self.format(body)
        if expr.pos[0] != body.pos[0]:
            formatted.extend(lines)
        else:
            formatted[-1] += lines[0]
            formatted.extend(lines[1:])

        formatted = formatted[0:1] + self._indent(formatted[1:], self.opts.indent)

        if expr.end_pos[0] != body.end_pos[0]:
            formatted.append(")")
        else:
            formatted[-1] += ")"

        return formatted

    def format_unary_op(self, expr):
        # TODO: enforce in type?
        assert len(expr.children) == 2

        op = self.format(expr.children[0])
        assert len(op) == 1

        lines = self.format(expr.children[1])
        lines[0] = op[0] + lines[0]
        return lines

    def _format_op(self, expr) -> List[str]:
        nodes = expr.children
        formatted = self.format(nodes[0])

        last = nodes[0]
        for next in nodes[1:]:
            lines = self.format(next)
            if last.end_pos[0] != next.pos[0]:
                formatted.extend(lines)
            else:
                formatted[-1] += " "
                formatted[-1] += lines[0]
                formatted.extend(lines[1:])
            last = next

        return formatted

    def format_binary_op(self, expr) -> List[str]:
        # TODO: enforce in type?
        assert len(expr.children) == 3
        return self._format_op(expr)

    def format_ternary_op(self, expr) -> List[str]:
        # TODO: enforce in type?
        assert len(expr.children) == 5
        return self._format_op(expr)

    def format_function(self, function):
        # TODO: enforce in type?
        assert len(function.children) >= 1

        name = self.format(function.children[0])
        assert len(name) == 1
        name = name[0]

        formatted = [f"{name}("]

        last = function.children[0]
        for i, child in enumerate(function.children[1:]):
            if child == BareWord(","):
                continue
            if i > 0:
                formatted[-1] += ","
            lines = self.format(child)
            if last.end_pos[0] != child.pos[0]:
                formatted.extend(lines)
            else:
                if i > 0:
                    formatted[-1] += " "
                formatted[-1] += lines[0]
                formatted.extend(lines[1:])
            last = child

        # indent any continuation lines, but we leave the closing paren dedented
        formatted = formatted[0:1] + self._indent(formatted[1:], self.opts.indent)

        if last.end_pos[0] != function.end_pos[0]:
            formatted.append(")")
        else:
            formatted[-1] += ")"

        return formatted
